
/*
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

// http://ceph.com/docs/master/dev/rbd-diff/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>


void main(int argc, char **argv) {

	FILE *fdiff;
	FILE *fimg;

	if(argc<3) {
		printf("Usage: %s <full image file> <diff file>\n", argv[0]);
		exit(0);
	}

	char *img=argv[1];
	fimg=fopen(img, "r+");
	if(fimg==NULL) {
		printf("error opening image file %s\n", img);
		exit(1);
	}

	char *diff=argv[2];
	fdiff=fopen(diff, "r");
	if(fdiff==NULL) {
		printf("error opening diff file %s\n", diff);
		exit(1);
	}
	char header[255];
	int n=fread(header, 1, 12, fdiff);
	if(n!=12) {
		printf("error reading diff header\n");
		exit(1);
	}
	if(strncmp(header, "rbd diff v1", 11)) {
		printf("unrecognised diff format\n");
		exit(1);
	}

	guint8 tag;
	int reading_metadata=1;
	printf("Reading metadata\n");
	while(reading_metadata) {
		int n=fread(&tag, 1, 1, fdiff);
		if(n!=1) {
			printf("error reading metadata\n");
			exit(1);
		}

		guint32 name_len;
		int rlen;
		#define MAX_LEN 1023
		char name[MAX_LEN+1];
		guint64 image_size;

		switch(tag) {
			case 'f':
				n=fread(&name_len, 1, sizeof(guint32), fdiff);
				rlen=name_len;
				if(rlen>MAX_LEN) rlen=MAX_LEN;
				n=fread(name, 1, rlen, fdiff);
				name[rlen]=0;
				printf("From snap: %s\n", name);
				if(name_len>MAX_LEN) fseek(fdiff, name_len-MAX_LEN, SEEK_CUR);
				break;
			case 't':
				n=fread(&name_len, 1, sizeof(guint32), fdiff);
				rlen=name_len;
				if(rlen>MAX_LEN) rlen=MAX_LEN;
				n=fread(name, 1, rlen, fdiff);
				name[rlen]=0;
				printf("To snap: %s\n", name);
				if(name_len>MAX_LEN) fseek(fdiff, name_len-MAX_LEN, SEEK_CUR);
				break;
			case 's':
				n=fread(&image_size, 1, sizeof(guint64), fdiff);
				printf("Image size: %lu (%luGB)\n", image_size, image_size/(1024*1024*1024));
				break;
			default:
				reading_metadata=0;
				printf("End of metadata\n");
				fseek(fdiff, -1, SEEK_CUR);
				break;
		}
	}

	char *rbuf=(char *)malloc(4096);
	char *zbuf=(char *)malloc(4096);
	memset((void *)zbuf, 0, 4096);

	int reading_data=1;
	while(reading_data) {
		int n=fread(&tag, 1, 1, fdiff);
		if(n!=1) {
			printf("error reading data\n");
			exit(1);
		}

		guint64 offset;
		guint64 length;
		int written;

		switch(tag) {
			case 'w':
				n=fread(&offset, 1, sizeof(guint64), fdiff);
				n=fread(&length, 1, sizeof(guint64), fdiff);
				printf("Data at offset %lu of length %lu\n", offset, length);
				//fseek(fdiff, length, SEEK_CUR);
				written=0;
				while(written<length) {
					int l;
					if(length-written>4096) l=4096;
					else l=length-written;
					n=fread(rbuf, 1, l, fdiff);
					if(n!=l) {
						printf("error reading diff file\n");
						exit(0);
					}
					n=fseek(fimg, offset+written, SEEK_SET);
					if(n) {
						printf("error seeking in image file\n");
						exit(0);
					}
					//printf("writing %d bytes to image\n", l);
					n=fwrite(rbuf, 1, l, fimg);
					written+=n;
					if(n==0) {
						printf("error writing to image\n");
						exit(0);
					}
				}
				break;
			case 'z':
				n=fread(&offset, 1, sizeof(guint64), fdiff);
				n=fread(&length, 1, sizeof(guint64), fdiff);
				printf("Data at offset %lu of length %lu\n", offset, length);

				int written=0;
				while(written<length) {
					int l;
					if(length-written>4096) l=4096;
					else l=length-written;
					n=fseek(fimg, offset+written, SEEK_SET);
					if(n) {
						printf("error seeking in image file\n");
						exit(0);
					}
					//printf("writing %d bytes to image\n", l);
					n=fwrite(zbuf, 1, l, fimg);
					written+=n;
					if(n==0) {
						printf("error writing to image\n");
						exit(0);
					}
				}
				break;
			case 'e':
				reading_data=0;
				printf("End of data\n");
				break;
			default:
				reading_data=0;
				printf("Unknown data encountered %d\n", tag);
				break;
		}
	}

	fclose(fimg);
	fclose(fdiff);
}

